"""Unit tests for reviewboard.site.managers.LocalSiteManager."""

from __future__ import annotations

from uuid import UUID, uuid4

import kgb
from django_assert_queries.testing import assert_queries
from django.db.models import Count, Q

from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class LocalSiteManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.site.managers.LocalSiteManager."""

    def test_get_stats_with_no_sites(self):
        """Testing LocalSiteManager.get_stats with no LocalSites"""
        uuids = self._pregenerate_cache_state_uuids(1)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 0,
            })

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 0,
            })

    def test_get_stats_with_sites(self):
        """Testing LocalSiteManager.get_stats with LocalSites"""
        uuids = self._pregenerate_cache_state_uuids(2)

        self.create_local_site(name='test-site-1')
        self.create_local_site(name='test-site-2')
        self.create_local_site(name='test-site-3',
                               public=True)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 2,
                'public_count': 1,
                'state_uuid': uuids[0],
                'total_count': 3,
            })

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 2,
                'public_count': 1,
                'state_uuid': uuids[0],
                'total_count': 3,
            })

    def test_get_stats_after_private_save(self):
        """Testing LocalSiteManager.get_stats after private LocalSite.save"""
        uuids = self._pregenerate_cache_state_uuids(2)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 0,
            })

        self.create_local_site(name='test-site-1')

        # The second query should hit cache.
        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 1,
                'public_count': 0,
                'state_uuid': uuids[1],
                'total_count': 1,
            })

    def test_get_stats_after_public_save(self):
        """Testing LocalSiteManager.get_stats after public LocalSite.save"""
        uuids = self._pregenerate_cache_state_uuids(2)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 0,
            })

        self.create_local_site(name='test-site-1',
                               public=True)

        # Cache should be invalidated.
        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 1,
                'state_uuid': uuids[1],
                'total_count': 1,
            })

    def test_get_stats_after_public_state_save(self):
        """Testing LocalSiteManager.get_stats after private -> public
        LocalSite.save
        """
        uuids = self._pregenerate_cache_state_uuids(2)

        local_site = self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 1,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 1,
            })

        local_site.public = True
        local_site.save(update_fields=('public',))

        # 2 queries:
        #
        # 1. Total LocalSite count
        # 2. Public LocalSite count
        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 1,
                'state_uuid': uuids[1],
                'total_count': 1,
            })

    def test_get_stats_after_delete(self):
        """Testing LocalSiteManager.get_stats after LocalSite.delete"""
        uuids = self._pregenerate_cache_state_uuids(2)

        local_site = self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 1,
                'public_count': 0,
                'state_uuid': uuids[0],
                'total_count': 1,
            })

        local_site.delete()

        # Cache should be invalidated.
        with assert_queries(queries):
            self.assertEqual(LocalSite.objects.get_stats(), {
                'private_count': 0,
                'public_count': 0,
                'state_uuid': uuids[1],
                'total_count': 0,
            })

    def test_get_local_site_acl_stats_with_local_site(self):
        """Testing LocalSiteManager.get_local_site_acl_stats with LocalSite"""
        uuids = self._pregenerate_cache_state_uuids(1)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')
        user3 = self.create_user(username='user3')
        user4 = self.create_user(username='user4')
        user5 = self.create_user(username='user5')

        local_site = self.create_local_site(name='test-site-1')
        local_site.users.add(user1, user2, user3)
        local_site.admins.add(user4, user5)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            stats = LocalSite.objects.get_local_site_acl_stats(local_site)

        self.assertEqual(stats, {
            'admin_count': 2,
            'user_count': 3,
            'public': False,
            'state_uuid': uuids[0],
        })

        # The second query should hit cache.
        with self.assertNumQueries(0):
            new_stats = LocalSite.objects.get_local_site_acl_stats(local_site)

        self.assertEqual(new_stats, stats)

    def test_get_local_site_acl_stats_with_local_site_id(self):
        """Testing LocalSiteManager.get_local_site_acl_stats with LocalSite ID
        """
        uuids = self._pregenerate_cache_state_uuids(1)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')
        user3 = self.create_user(username='user3')
        user4 = self.create_user(username='user4')
        user5 = self.create_user(username='user5')

        local_site = self.create_local_site(name='test-site-1')
        local_site.users.add(user1, user2, user3)
        local_site.admins.add(user4, user5)

        # 3 queries:
        #
        # 1. Local Site instance
        # 2. User count
        # 3. Admin count
        queries = [
            {
                'model': LocalSite,
                'only_fields': {'public'},
                'where': Q(pk=local_site.pk),
            },
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            stats = LocalSite.objects.get_local_site_acl_stats(local_site.pk)

        self.assertEqual(stats, {
            'admin_count': 2,
            'user_count': 3,
            'public': False,
            'state_uuid': uuids[0],
        })

        # The second query should hit cache.
        with self.assertNumQueries(0):
            new_stats = LocalSite.objects.get_local_site_acl_stats(
                local_site.pk)

        self.assertEqual(new_stats, stats)

    def test_get_local_site_acl_stats_after_save(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after changing
        public state
        """
        uuids = self._pregenerate_cache_state_uuids(2)

        local_site = self.create_local_site(name='test-site-1')

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.public = True
        local_site.save()

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': True,
                    'state_uuid': uuids[1],
                })

    def test_get_local_site_acl_stats_after_save_update_fields_public(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        LocalSite.save(update_fields=['public'])
        """
        uuids = self._pregenerate_cache_state_uuids(2)

        local_site = self.create_local_site(name='test-site-1',
                                            public=True)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': True,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.public = False
        local_site.save(update_fields=('public',))

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[1],
                })

    def test_get_local_site_acl_stats_after_save_update_fields(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        LocalSite.save(update_fields=) without public
        """
        uuids = self._pregenerate_cache_state_uuids(1)

        local_site = self.create_local_site(name='test-site-1')

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            stats = LocalSite.objects.get_local_site_acl_stats(local_site)

        self.assertEqual(stats, {
            'admin_count': 0,
            'user_count': 0,
            'public': False,
            'state_uuid': uuids[0],
        })

        # This should NOT invalidate the cache.
        local_site.save(update_fields=('name',))

        with self.assertNumQueries(0):
            new_stats = LocalSite.objects.get_local_site_acl_stats(local_site)

        self.assertEqual(new_stats, stats)

    def test_get_local_site_acl_stats_after_delete(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        LocalSite.delete
        """
        uuids = self._pregenerate_cache_state_uuids(1)

        local_site = self.create_local_site(name='test-site-1')
        local_site_id = local_site.pk

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.delete()

        # Any subsequent fetches should return None (which will still be
        # cached).
        queries = [
            {
                'model': LocalSite,
                'only_fields': {'public'},
                'where': Q(pk=local_site_id),
            },
        ]

        with assert_queries(queries):
            self.assertIsNone(
                LocalSite.objects.get_local_site_acl_stats(local_site_id))

        with self.assertNumQueries(0):
            self.assertIsNone(
                LocalSite.objects.get_local_site_acl_stats(local_site_id))

    def test_get_local_site_acl_stats_after_add_user(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        adding to LocalSite.users
        """
        uuids = self._pregenerate_cache_state_uuids(3)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.users.add(user1)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 1,
                    'public': False,
                    'state_uuid': uuids[1],
                })

        # Now add it on the reverse relation. This should invalidate the
        # cache.
        user2.local_site.add(local_site)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 2,
                    'public': False,
                    'state_uuid': uuids[2],
                })

    def test_get_local_site_acl_stats_after_remove_user(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        removing from LocalSite.users
        """
        uuids = self._pregenerate_cache_state_uuids(3)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')
        local_site.users.add(user1, user2)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 2,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.users.remove(user1)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 1,
                    'public': False,
                    'state_uuid': uuids[1],
                })

        # Now remove on the reverse relation. This should invalidate the
        # cache.
        user2.local_site.remove(local_site)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[2],
                })

    def test_get_local_site_acl_stats_after_clear_user(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        clearing LocalSite.users
        """
        uuids = self._pregenerate_cache_state_uuids(2)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')
        local_site.users.add(user1, user2)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 2,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.users.clear()

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[1],
                })

    def test_get_local_site_acl_stats_after_add_admin(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        adding to LocalSite.admins
        """
        uuids = self._pregenerate_cache_state_uuids(3)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.admins.add(user1)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 1,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[1],
                })

        # Now add it on the reverse relation. This should invalidate the
        # cache.
        user2.local_site_admins.add(local_site)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 2,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[2],
                })

    def test_get_local_site_acl_stats_after_remove_admin(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        removing from LocalSite.admins
        """
        uuids = self._pregenerate_cache_state_uuids(3)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')
        local_site.admins.add(user1, user2)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 2,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.admins.remove(user1)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 1,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[1],
                })

        # Now remove on the reverse relation. This should invalidate the
        # cache.
        user2.local_site_admins.remove(local_site)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[2],
                })

    def test_get_local_site_acl_stats_after_clear_admins(self):
        """Testing LocalSiteManager.get_local_site_acl_stats after
        clearing LocalSite.admins
        """
        uuids = self._pregenerate_cache_state_uuids(2)

        user1 = self.create_user(username='user1')
        user2 = self.create_user(username='user2')

        local_site = self.create_local_site(name='test-site-1')
        local_site.admins.add(user1, user2)

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        queries = [
            {
                'model': LocalSite.users.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
            {
                'model': LocalSite.admins.through,
                'annotations': {'__count': Count('*')},
                'where': Q(localsite=local_site.pk),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 2,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[0],
                })

        # This should invalidate the cache.
        local_site.admins.clear()

        # 2 queries:
        #
        # 1. User count
        # 2. Admin count
        with assert_queries(queries):
            self.assertEqual(
                LocalSite.objects.get_local_site_acl_stats(local_site),
                {
                    'admin_count': 0,
                    'user_count': 0,
                    'public': False,
                    'state_uuid': uuids[1],
                })

    def test_has_local_sites_with_no_sites(self):
        """Testing LocalSiteManager.has_local_sites with no LocalSites"""
        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertFalse(LocalSite.objects.has_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertFalse(LocalSite.objects.has_local_sites())

    def test_has_local_sites_with_public_sites(self):
        """Testing LocalSiteManager.has_local_sites with public LocalSites"""
        self.create_local_site(name='test-site-1',
                               public=True)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertTrue(LocalSite.objects.has_local_sites())

    def test_has_local_sites_with_private_sites(self):
        """Testing LocalSiteManager.has_local_sites with private LocalSites"""
        self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertTrue(LocalSite.objects.has_local_sites())

    def test_has_public_local_sites_with_no_public_sites(self):
        """Testing LocalSiteManager.has_public_local_sites with no public
        LocalSites
        """
        self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertFalse(LocalSite.objects.has_public_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertFalse(LocalSite.objects.has_public_local_sites())

    def test_has_public_local_sites_with_public_sites(self):
        """Testing LocalSiteManager.has_public_local_sites with public
        LocalSites
        """
        self.create_local_site(name='test-site-1',
                               public=True)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_public_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertTrue(LocalSite.objects.has_public_local_sites())

    def test_has_private_local_sites_with_no_private_sites(self):
        """Testing LocalSiteManager.has_private_local_sites with no private
        LocalSites
        """
        self.create_local_site(name='test-site-1',
                               public=True)

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertFalse(LocalSite.objects.has_private_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertFalse(LocalSite.objects.has_private_local_sites())

    def test_has_private_local_sites_with_private_sites(self):
        """Testing LocalSiteManager.has_private_local_sites with private
        LocalSites
        """
        self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_private_local_sites())

        # The second query should hit cache.
        with self.assertNumQueries(0):
            self.assertTrue(LocalSite.objects.has_private_local_sites())

    def test_invalidate_stats_cache(self):
        """Testing LocalSiteManager.invalidate_stats_cache"""
        self.create_local_site(name='test-site-1')

        # 1 query:
        #
        # 1. Total and public LocalSite counts
        queries = [
            {
                'model': LocalSite,
                'annotations': {
                    'total': Count('*'),
                    'public_count': Count('public', filter=Q(public=True)),
                },
            },
        ]

        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_local_sites())

        LocalSite.objects.invalidate_stats_cache()

        # 2 queries:
        #
        # 1. Total LocalSite count
        # 2. Public LocalSite count
        with assert_queries(queries):
            self.assertTrue(LocalSite.objects.has_local_sites())

    def _pregenerate_cache_state_uuids(self, count):
        """Pre-generate a list of UUIDs for state_uuid cache information.

        This will generate a list, ensure :py:func:`uuid.uuid4` returns them
        in the order called, and returns them for reference.

        Args:
            count (int):
                The number of UUIDs to pre-generate.

        Returns:
            list of str:
            The list of pre-generated UUIDs.
        """
        uuids = [
            '00000000-0000-0000-0000-%012d' % _i
            for _i in range(count)
        ]

        self.spy_on(uuid4, op=kgb.SpyOpReturnInOrder([
            UUID(_uuid)
            for _uuid in uuids
        ]))

        return uuids



"""Unit tests for extension hook sandboxing."""

from django.contrib.auth.models import User
from django.template import Context, Template
from django.test.client import RequestFactory
from djblets.extensions.models import RegisteredExtension

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.extensions.base import Extension
from reviewboard.extensions.hooks import (CommentDetailDisplayHook,
                                          DiffViewerActionHook,
                                          FileDiffACLHook,
                                          HeaderActionHook,
                                          HeaderDropdownActionHook,
                                          NavigationBarHook,
                                          ReviewRequestActionHook,
                                          ReviewRequestApprovalHook,
                                          ReviewRequestDropdownActionHook,
                                          ReviewRequestFieldSetsHook,
                                          UserInfoboxHook)
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.reviews.fields import (BaseReviewRequestField,
                                        BaseReviewRequestFieldSet)
from reviewboard.reviews.models import ReviewRequest
from reviewboard.site.urlresolvers import local_site_reverse


class SandboxExtension(Extension):
    registration = RegisteredExtension()
    metadata = {
        'Name': 'Sandbox Extension',
    }
    id = 'reviewboard.extensions.tests.SandboxExtension'


class SandboxFileDiffACLTestHook(FileDiffACLHook):
    def is_accessible(self, diffset, filediff, user):
        raise Exception


class SandboxReviewRequestApprovalTestHook(ReviewRequestApprovalHook):
    def is_approved(self, review_request, prev_approved, prev_failure):
        raise Exception


class SandboxNavigationBarTestHook(NavigationBarHook):
    def get_entries(self, context):
        raise Exception


class SandboxDiffViewerActionTestHook(DiffViewerActionHook):
    def get_actions(self, context):
        raise Exception


class SandboxHeaderActionTestHook(HeaderActionHook):
    def get_actions(self, context):
        raise Exception


class SandboxHeaderDropdownActionTestHook(HeaderDropdownActionHook):
    def get_actions(self, context):
        raise Exception


class SandboxReviewRequestActionTestHook(ReviewRequestActionHook):
    def get_actions(self, context):
        raise Exception


class SandboxReviewRequestDropdownActionTestHook(
        ReviewRequestDropdownActionHook):
    def get_actions(self, context):
        raise Exception


class SandboxCommentDetailDisplayTestHook(CommentDetailDisplayHook):
    def render_review_comment_detail(self, comment):
        raise Exception

    def render_email_comment_detail(self, comment, is_html):
        raise Exception


class SandboxBaseReviewRequestTestShouldRenderField(BaseReviewRequestField):
    field_id = 'should_render'

    def should_render(self, value):
        raise Exception


class SandboxBaseReviewRequestTestInitField(BaseReviewRequestField):
    field_id = 'init_field'

    def __init__(self, review_request_details):
        raise Exception


class SandboxUserInfoboxHook(UserInfoboxHook):
    def get_etag_data(self, user, request, local_site):
        raise Exception

    def render(self, user, request, local_site):
        raise Exception


class TestIsEmptyField(BaseReviewRequestField):
    field_id = 'is_empty'


class TestInitField(BaseReviewRequestField):
    field_id = 'test_init'


class TestInitFieldset(BaseReviewRequestFieldSet):
    fieldset_id = 'test_init'
    field_classes = [SandboxBaseReviewRequestTestInitField]


class TestShouldRenderFieldset(BaseReviewRequestFieldSet):
    fieldset_id = 'test_should_render'
    field_classes = [SandboxBaseReviewRequestTestShouldRenderField]


class BaseReviewRequestTestIsEmptyFieldset(BaseReviewRequestFieldSet):
    fieldset_id = 'is_empty'
    field_classes = [TestIsEmptyField]

    @classmethod
    def is_empty(cls):
        raise Exception


class BaseReviewRequestTestInitFieldset(BaseReviewRequestFieldSet):
    fieldset_id = 'init_fieldset'
    field_classes = [TestInitField]

    def __init__(self, review_request_details):
        raise Exception


class SandboxTests(BaseExtensionHookTestCase):
    """Testing extension sandboxing."""

    extension_class = SandboxExtension
    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(SandboxTests, self).setUp()

        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='reviewboard',
                                             email='reviewboard@example.com',
                                             password='password')

    def test_filediff_acl_sandbox(self):
        """Testing FileDiffACLHook.is_accessible with raised exception
        """
        SandboxFileDiffACLTestHook(extension=self.extension)
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        author_name = review_request.submitter.get_full_name()
        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=author_name),

        review_request._are_diffs_accessible_by(self.user)

    def test_is_approved_sandbox(self):
        """Testing ReviewRequestApprovalHook.is_approved with raised exception
        """
        SandboxReviewRequestApprovalTestHook(extension=self.extension)
        review = ReviewRequest()
        review._calculate_approval()

    def test_get_entries(self):
        """Testing NavigationBarHook.get_entries with raised exception"""
        entry = {
            'label': 'Test get_entries Function',
            'url': '/dashboard/',
        }

        SandboxNavigationBarTestHook(extension=self.extension, entries=[entry])

        context = Context({})

        t = Template(
            "{% load rb_extensions %}"
            "{% navigation_bar_hooks %}")

        t.render(context).strip()

    def test_render_review_comment_details(self):
        """Testing CommentDetailDisplayHook when render_review_comment_detail
        throws an error
        """
        SandboxCommentDetailDisplayTestHook(extension=self.extension)

        context = Context({'comment': 'this is a comment'})

        t = Template(
            "{% load rb_extensions %}"
            "{% comment_detail_display_hook comment 'review'%}")

        t.render(context).strip()

    def test_email_review_comment_details(self):
        """Testing CommentDetailDisplayHook.render_email_comment_detail with
        raised exception
        """
        SandboxCommentDetailDisplayTestHook(extension=self.extension)

        context = Context({'comment': 'this is a comment'})

        t = Template(
            "{% load rb_extensions %}"
            "{% comment_detail_display_hook comment 'html-email'%}")

        t.render(context).strip()

    def test_action_hooks_diff_viewer_hook(self):
        """Testing DiffViewerActionHook.get_actions with raised exception"""
        deprecationMessage = (
            'DiffViewerActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension '
            '"reviewboard.extensions.tests.test_sandbox_hooks.'
            'SandboxExtension" will need to be updated to derive actions '
            'from reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecationMessage):
            SandboxDiffViewerActionTestHook(extension=self.extension)

        context = Context({
            'comment': 'this is a comment',
            'request': self.factory.get('/test/'),
        })

        template = Template(
            '{% load actions %}'
            '{% actions_html "review-request" %}')

        template.render(context)

    def test_action_hooks_review_request_hook(self):
        """Testing ReviewRequestActionHook.get_actions with raised exception"""
        deprecationMessage = (
            'ReviewRequestActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension '
            '"reviewboard.extensions.tests.test_sandbox_hooks.'
            'SandboxExtension" will need to be updated to derive actions '
            'from reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecationMessage):
            SandboxReviewRequestActionTestHook(extension=self.extension)

        context = Context({
            'comment': 'this is a comment',
            'request': self.factory.get('/test/'),
        })

        template = Template(
            '{% load actions %}'
            '{% actions_html "review-request" %}')

        template.render(context)

    def test_action_hooks_review_request_dropdown_hook(self):
        """Testing ReviewRequestDropdownActionHook.get_actions with raised
        exception
        """
        deprecationMessage = (
            'ReviewRequestDropdownActionHook is deprecated and will be '
            'removed in Review Board 9. Your extension '
            '"reviewboard.extensions.tests.test_sandbox_hooks.'
            'SandboxExtension" will need to be updated to derive actions '
            'from reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecationMessage):
            SandboxReviewRequestDropdownActionTestHook(
                extension=self.extension)

        context = Context({
            'comment': 'this is a comment',
            'request': self.factory.get('/test/'),
        })

        template = Template(
            '{% load actions %}'
            '{% actions_html "review-request" %}')

        template.render(context)

    def test_is_empty_review_request_fieldset(self):
        """Testing ReviewRequestFieldSetsHook when
        BaseReviewRequestField.is_empty raises exception
        """
        fieldset = [BaseReviewRequestTestIsEmptyFieldset]
        ReviewRequestFieldSetsHook(extension=self.extension,
                                   fieldsets=fieldset)

        review = ReviewRequest()

        request = self.factory.get('test')
        request.user = self.user
        context = Context({
            'review_request_details': review,
            'request': request
        })

        t = Template(
            "{% load reviewtags %}"
            "{% for_review_request_fieldset review_request_details %}"
            "{% end_for_review_request_fieldset %}")

        t.render(context).strip()

    def test_field_cls_review_request_field(self):
        """Testing ReviewRequestFieldSetsHook when
        BaseReviewRequestField.__init__ raises exception
        """
        fieldset = [TestInitFieldset]
        ReviewRequestFieldSetsHook(extension=self.extension,
                                   fieldsets=fieldset)

        review = ReviewRequest()
        context = Context({
            'review_request_details': review,
            'fieldset': TestInitFieldset
        })

        t = Template(
            "{% load reviewtags %}"
            "{% for_review_request_field review_request_details 'test_init' %}"
            "{% end_for_review_request_field %}")

        t.render(context).strip()

    def test_fieldset_cls_review_request_fieldset(self):
        """Testing ReviewRequestFieldSetsHook when
        ReviewRequestFieldset.__init__ raises exception
        """
        fieldset = [BaseReviewRequestTestInitFieldset]
        ReviewRequestFieldSetsHook(extension=self.extension,
                                   fieldsets=fieldset)

        review = ReviewRequest()
        request = self.factory.get('test')
        request.user = self.user
        context = Context({
            'review_request_details': review,
            'request': request
        })

        t = Template(
            "{% load reviewtags %}"
            "{% for_review_request_fieldset review_request_details %}"
            "{% end_for_review_request_fieldset %}")

        t.render(context).strip()

    def test_should_render_review_request_field(self):
        """Testing ReviewRequestFieldSetsHook when
        ReviewRequestFieldset.should_render raises exception
        """
        fieldset = [TestShouldRenderFieldset]
        ReviewRequestFieldSetsHook(extension=self.extension,
                                   fieldsets=fieldset)

        review = ReviewRequest()
        context = Context({
            'review_request_details': review,
            'fieldset': TestShouldRenderFieldset
        })

        t = Template(
            "{% load reviewtags %}"
            "{% for_review_request_field review_request_details"
            " 'test_should_render' %}"
            "{% end_for_review_request_field %}")

        t.render(context).strip()

    def test_user_infobox_hook(self):
        """Testing UserInfoboxHook with methods raising exceptions"""
        SandboxUserInfoboxHook(self.extension, 'template.html')

        self.client.get(
            local_site_reverse('user-infobox', kwargs={
                'username': self.user.username,
            }))



"""Utilities for managing users with SSO.

Version Added:
    5.0
"""

import re

from django.contrib.auth.models import User
from django.db.models import Q

from reviewboard.accounts.sso.errors import InvalidUsernameError


# This matches what's used in Review Board URLs.
INVALID_USERNAME_CHARS_RE = re.compile(r'[^\w.@+-]+')


def find_suggested_username(email):
    """Return the suggested username for a given e-mail address.

    Version Added:
        5.0

    Args:
        email (str):
            The user's e-mail address.

    Returns:
        str:
        The suggested username.

    Raises:
        reviewboard.accounts.sso.errors.InvalidUsernameError:
            The suggested username does not work with Review Board's rules.
    """
    # Normalize the username, factoring in everything before the '@' (if any)
    # and converting any non-alphanumeric characters to dashes. Then we'll
    # sanity-check that it meets our username criteria.
    norm_user_id = INVALID_USERNAME_CHARS_RE.sub('-', email.split('@')[0])

    if (norm_user_id.startswith('-') or
        norm_user_id.endswith('-') or
        '--' in norm_user_id):
        # This is an invalid username.
        raise InvalidUsernameError('Invalid SSO username "%s"' % norm_user_id)

    return norm_user_id


def find_user_for_sso_user_id(username, email, alternate_username):
    """Find a matching user for SSO login.

    Version Added:
        5.0

    Args:
        username (str):
            The username, if available. May be ``None``.

        email (str):
            The user's e-mail address.

        alternate_username (str):
            An alternate username to try, if ``username`` is ``None``. This
            probably comes from :py:func:`find_suggested_username`.

    Returns:
        django.contrib.auth.models.User:
        The user object, if one exists.
    """
    q = Q()

    if username:
        q |= Q(username=username)

    q |= Q(email=email) | Q(username=alternate_username)

    # Fetch all candidates first to save on DB queries.
    candidate_users = list(User.objects.filter(q))

    # First try by username. This will be present if the username has been
    # specified in the SSO provider.
    if username:
        for user in candidate_users:
            if user.username == username:
                return user

    # Next see if we have a user with a matching e-mail address.
    for user in candidate_users:
        if user.email == email:
            return user

    # Finally, try a computed username from the e-mail address.
    for user in candidate_users:
        if user.username == alternate_username:
            return user

    return None



"""Models for Local Sites."""

from __future__ import annotations

from enum import Enum
from typing import ClassVar, Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField
from housekeeping.functions import deprecate_non_keyword_only_args
from typing_extensions import Literal

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.site.managers import LocalSiteManager

if TYPE_CHECKING:
    from typing import Final, TypeAlias

    from django.contrib.auth.models import AnonymousUser


class _LocalSiteConstants(Enum):
    """Constants for special LocalSite queries.

    Version Added:
        6.0:
        This replaces an older ALL-specific object, to aid in typing.
    """

    ALL = '<ALL>'


class LocalSite(models.Model):
    """A division within a Review Board installation.

    This allows the creation of independent, isolated divisions within a given
    server. Users can be designated as members of a LocalSite, and optionally
    as admins (which allows them to manipulate the repositories, groups and
    users in the site).

    Pretty much every other model in this module can all be assigned to a
    single LocalSite, at which point only members will be able to see or
    manipulate these objects. Access control is performed at every level, and
    consistency is enforced through a liberal sprinkling of assertions and unit
    tests.
    """

    #: A sentinel indicating all LocalSites.
    #:
    #: This is supported by some functions to perform a query against all
    #: LocalSites, as opposed to either a single LocalSite instance or the
    #: global site through ``None``.
    #:
    #: Check the documentation for a function's argument to determine whether
    #: this is supported.
    #:
    #: Note that the value is considered opaque. It can be compared directly
    #: using ``is``.
    #:
    #: Version Added:
    #:     5.0
    ALL: Final[Literal[_LocalSiteConstants.ALL]] = _LocalSiteConstants.ALL

    name = models.SlugField(
        _('name'),
        max_length=32,
        blank=False,
        unique=True,
    )

    public = models.BooleanField(
        default=False,
        db_index=True,
        help_text=_(
            'Allow people outside the team to access and post review '
            'requests and reviews.'
        ),
    )

    users: models.ManyToManyField[User, User] = models.ManyToManyField(
        User,
        blank=True,
        related_name='local_site',
    )

    admins: models.ManyToManyField[User, User] = models.ManyToManyField(
        User,
        blank=True,
        related_name='local_site_admins',
    )

    extra_data = JSONField(null=True)

    objects: ClassVar[LocalSiteManager] = LocalSiteManager()

    def is_accessible_by(
        self,
        user: AnonymousUser | User,
    ) -> bool:
        """Return whether or not the user has access to this LocalSite.

        This checks that the user is logged in, and that they're listed in the
        'users' field.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            ``True`` if the user has access. ``False`` if they do not.
        """
        return (self.public or
                (user.is_authenticated and
                 (user.is_staff or self.users.filter(pk=user.pk).exists())))

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def is_mutable_by(
        self,
        user: AnonymousUser | User,
        *,
        perm: str = 'site.change_localsite',
    ) -> bool:
        """Return whether or not a user can modify settings in a LocalSite.

        This checks that the user is either staff with the proper permissions,
        or that they're listed in the 'admins' field.

        By default, this is checking whether the LocalSite itself can be
        modified, but a different permission can be passed to check for
        another object.

        Version Changed:
            8.0:
            ``perm`` must now be provided as a keyword-only argument.
            Support for providing these as positional arguments is
            deprecated and will be removed in Review Board 10.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check.

            perm (str, optional):
                The explicit permission to check.

                Defaults to ``site.change_localsite``.

        Returns:
            bool:
            ``True`` if the user can modify the LocalSite settings.
            ``False`` if they cannot.
        """
        return user.has_perm(perm) or self.admins.filter(pk=user.pk).exists()

    def __str__(self) -> str:
        """Return the string form of the Local Site.

        This will return :py:attr:`name` directly.

        Returns:
            str:
            The Local Site name.
        """
        return self.name

    class Meta:
        db_table = 'site_localsite'
        verbose_name = _('Local Site')
        verbose_name_plural = _('Local Sites')


#: A filter value for matching a LocalSite, all LocalSites, or None.
#:
#: Version Added:
#:     5.0.7
AnyOrAllLocalSites: TypeAlias = Union[Optional[LocalSite],
                                      Literal[_LocalSiteConstants.ALL]]



"""Base class for file attachment comment resources."""

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.timesince import timesince
from djblets.util.decorators import augment_method_from
from djblets.webapi.fields import ResourceFieldType, StringFieldType

from reviewboard.reviews.models import FileAttachmentComment
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_comment import BaseCommentResource


class BaseFileAttachmentCommentResource(BaseCommentResource):
    """Base class for file attachment comment resources.

    Provides common fields and functionality for all file attachment comment
    resources. The list of comments cannot be modified from this resource.
    """

    added_in = '1.6'

    model = FileAttachmentComment
    name = 'file_attachment_comment'
    fields = dict({
        'diff_against_file_attachment': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.file_attachment.'
                        'FileAttachmentResource',
            'description': 'The file changes were made against in a diff.',
            'added_in': '2.0',
        },
        'file_attachment': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.file_attachment.'
                        'FileAttachmentResource',
            'description': 'The file the comment was made on.',
        },
        'link_text': {
            'type': StringFieldType,
            'description': 'The text used to describe a link to the file. '
                           'This may differ depending on the comment.',
            'added_in': '1.7.10',
        },
        'review_url': {
            'type': StringFieldType,
            'description': 'The URL to the review UI for the comment on this '
                           'file attachment.',
            'added_in': '1.7.10',
        },
        'thumbnail_html': {
            'type': StringFieldType,
            'description': 'The HTML representing a thumbnail, if any, for '
                           'this comment.',
            'added_in': '1.7.10',
        },
    }, **BaseCommentResource.fields)

    uri_object_key = 'comment_id'
    allowed_methods = ('GET',)

    def get_queryset(self, request, review_request_id=None, *args, **kwargs):
        """Return a queryset for FileAttachmentComment models.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review_request_id (int, optional):
                The review request ID used to filter the results. If set,
                only comments from the given review request that are public
                or owned by the requesting user will be included.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for FileAttachmentComment models.
        """
        q = Q(review__isnull=False)

        if review_request_id is not None:
            try:
                review_request = resources.review_request.get_object(
                    request, review_request_id=review_request_id,
                    *args, **kwargs)
            except ObjectDoesNotExist:
                raise self.model.DoesNotExist

            query = (Q(file_attachment__review_request=review_request) |
                     Q(file_attachment__inactive_review_request=review_request))

            draft = review_request.get_draft(request.user)

            if draft is not None:
                query |= (Q(file_attachment__drafts=draft) |
                          Q(file_attachment__inactive_drafts=draft))

            q &= query

        return self.model.objects.filter(q)

    def serialize_link_text_field(self, obj, **kwargs):
        return obj.get_link_text()

    def serialize_public_field(self, obj, **kwargs):
        return obj.review.get().public

    def serialize_review_url_field(self, obj, **kwargs):
        return obj.get_review_url()

    def serialize_thumbnail_html_field(self, obj, **kwargs):
        return obj.thumbnail

    def serialize_timesince_field(self, obj, **kwargs):
        return timesince(obj.timestamp)

    def serialize_user_field(self, obj, **kwargs):
        return obj.review.get().user

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Returns information on the comment.

        This contains the comment text, time the comment was made,
        and the file the comment was made on, amongst other information.
        """
        pass



"""Unit tests for reviewboard.reviews.builtin_fields."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Q
from django.test.client import RequestFactory
from django.urls import resolve
from django.utils.safestring import SafeText

from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.builtin_fields import (BugsField,
                                                CommitListField,
                                                FileAttachmentsField,
                                                ReviewRequestPageDataMixin)
from reviewboard.reviews.detail import ReviewRequestPageData
from reviewboard.reviews.fields import BaseReviewRequestField
from reviewboard.reviews.models import ReviewRequestDraft
from reviewboard.testing.testcase import TestCase

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict


class FieldsTestCase(TestCase):
    """Base test case for built-in fields."""

    field_cls: type[BaseReviewRequestField]

    def make_field(self, review_request):
        """Return an instance of the field to test with.

        The field will be populated with all review request page state.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being tested.

        Returns:
            reviewboard.reviews.fields.BaseReviewRequestField:
            The resulting field instance.
        """
        request = self.build_review_request_get(review_request)

        kwargs: KwargsDict = {
            'request': request,
        }

        if issubclass(self.field_cls, ReviewRequestPageDataMixin):
            data = ReviewRequestPageData(review_request, request)
            data.query_data_pre_etag()
            data.query_data_post_etag()
            kwargs['data'] = data

        return self.field_cls(review_request, **kwargs)

    def build_review_request_get(self, review_request):
        """Return an HTTP GET request for the review request.

        This will return a new HTTP request to a review request's page,
        containing a resolver match, without actually fetching the page.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being tested.

        Returns:
            django.http.HttpRequest:
            The request for the review request detail page.
        """
        url = review_request.get_absolute_url()
        request = RequestFactory().get(url)
        request.user = AnonymousUser()
        request.resolver_match = resolve(url)

        return request


class CommitListFieldTests(FieldsTestCase):
    """Unit tests for CommitListField."""

    field_cls = CommitListField
    fixtures = ['test_scmtools', 'test_users']

    def test_should_render_history_review_request(self):
        """Testing CommitListField.should_render with a review request created
        with history
        """
        review_request = self.create_review_request(create_repository=True,
                                                    create_with_history=True)
        request = self.build_review_request_get(review_request)
        self.create_diffset(review_request)
        field = CommitListField(review_request, request=request)

        self.assertTrue(field.should_render)

    def test_should_render_history_draft(self):
        """Testing CommitListField.should_render with a draft of a review
        request created with history
        """
        review_request = self.create_review_request(create_repository=True,
                                                    create_with_history=True)
        self.create_diffset(review_request, draft=True)
        request = self.build_review_request_get(review_request)
        field = CommitListField(review_request.get_draft(), request=request)

        self.assertTrue(field.should_render)

    def test_should_render_no_history_review_request(self):
        """Testing CommitListField.should_render with a review request created
        without history
        """
        review_request = self.create_review_request()
        request = self.build_review_request_get(review_request)
        field = CommitListField(review_request, request=request)

        self.assertFalse(field.should_render)

    def test_should_render_no_history_draft(self):
        """Testing CommitListField.should_render with a draft of a review
        request created without history
        """
        review_request = self.create_review_request()
        draft = ReviewRequestDraft.create(review_request)
        request = self.build_review_request_get(review_request)
        field = CommitListField(draft, request=request)

        self.assertFalse(field.should_render)

    def test_should_render_with_no_value(self):
        """Testing CommitListField.should_render with no value"""
        review_request = self.create_review_request(create_with_history=True)
        draft = ReviewRequestDraft.create(review_request)
        request = self.build_review_request_get(review_request)
        field = CommitListField(draft, request=request)

        self.assertFalse(field.should_render)

    def test_can_record_change_entry_history_review_request(self):
        """Testing CommitListField.can_record_change_entry with a review
        request created with history
        """
        review_request = self.create_review_request(create_with_history=True)
        field = CommitListField(review_request)

        self.assertTrue(field.can_record_change_entry)

    def test_can_record_change_entry_history_draft(self):
        """Testing CommitListField.can_record_change_entry with a draft of a
        review request created with history
        """
        review_request = self.create_review_request(create_with_history=True)
        draft = ReviewRequestDraft.create(review_request)
        field = CommitListField(draft)

        self.assertTrue(field.can_record_change_entry)

    def test_can_record_change_entry_no_history_review_request(self):
        """Testing CommitListField.can_record_change_entry with a review
        request created without history
        """
        review_request = self.create_review_request()
        field = CommitListField(review_request)

        self.assertFalse(field.can_record_change_entry)

    def test_can_record_change_entry_no_history_draft(self):
        """Testing CommitListField.can_record_change_entry with a draft of a
        review request created without history
        """
        review_request = self.create_review_request()
        draft = ReviewRequestDraft.create(review_request)
        field = CommitListField(draft)

        self.assertFalse(field.can_record_change_entry)

    def test_render_value(self):
        """Testing CommitListField.render_value"""
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        author_name = review_request.submitter.get_full_name()
        self.assertEqual(author_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=author_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=author_name)

        field = self.make_field(review_request)
        result = field.render_value(field.load_value(review_request))

        self.assertHTMLEqual(
            result,
            """
            <div class="rb-c-review-request-field-tabular rb-c-commit-list">
             <table class="rb-c-review-request-field-tabular__data">
              <thead>
               <tr>
                <th class="rb-c-commit-list__column-summary">Summary</th>
                <th class="rb-c-commit-list__column-id">ID</th>
               </tr>
              </thead>
              <tbody>
               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 1</div>
                </td>
                <td class="rb-c-commit-list__id" title="r1">r1</td>
               </tr>

               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 2</div>
                </td>
                <td class="rb-c-commit-list__id" title="r2">r2</td>
               </tr>
              </tbody>
             </table>
            </div>
            """)

    def test_render_value_with_author(self):
        """Testing CommitListField.render_value with an author that differs
        from the review request submitter
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()
        self.assertEqual(submitter_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name='Example Author')
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=submitter_name)

        field = self.make_field(review_request)
        result = field.render_value(field.load_value(review_request))

        self.assertHTMLEqual(
            result,
            """
            <div class="rb-c-review-request-field-tabular rb-c-commit-list">
             <table class="rb-c-review-request-field-tabular__data">
              <thead>
               <tr>
                <th class="rb-c-commit-list__column-summary">Summary</th>
                <th class="rb-c-commit-list__column-id">ID</th>
                <th class="rb-c-commit-list__column-author">Author</th>
               </tr>
              </thead>
              <tbody>
               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 1</div>
                </td>
                <td class="rb-c-commit-list__id" title="r1">r1</td>
                <td class="rb-c-commit-list__author">Example Author</td>
               </tr>

               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 2</div>
                </td>
                <td class="rb-c-commit-list__id" title="r2">r2</td>
                <td class="rb-c-commit-list__author">Doc Dwarf</td>
               </tr>
              </tbody>
             </table>
            </div>
            """)

    def test_render_value_with_collapse(self):
        """Testing CommitListField.render_value with a multi-line commit
        message
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        author_name = review_request.submitter.get_full_name()
        self.assertEqual(author_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=author_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2\n'
                                              'Longer message\n',
                               author_name=author_name)

        field = self.make_field(review_request)
        result = field.render_value(field.load_value(review_request))

        self.assertHTMLEqual(
            result,
            """
            <div class="rb-c-review-request-field-tabular rb-c-commit-list">
             <table class="rb-c-review-request-field-tabular__data">
              <thead>
               <tr>
                <th class="rb-c-commit-list__column-summary">Summary</th>
                <th class="rb-c-commit-list__column-id">ID</th>
               </tr>
              </thead>
              <tbody>
               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 1</div>
                </td>
                <td class="rb-c-commit-list__id" title="r1">r1</td>
               </tr>

               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <details>
                  <summary class="rb-c-commit-list__message-summary">Commit
                   message 2</summary>
                  <div class="rb-c-commit-list__message-body">Longer
                   message</div>
                 </details>
                </td>
                <td class="rb-c-commit-list__id" title="r2">r2</td>
               </tr>
              </tbody>
             </table>
            </div>
            """)

    def test_render_value_with_collapse_and_author(self):
        """Testing CommitListField.render_value with an author that differs
        from the review request submitter and a multi-line commit message
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()
        self.assertEqual(submitter_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name='Example Author')
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2\n'
                                              'Longer message\n',
                               author_name=submitter_name)

        field = self.make_field(review_request)
        result = field.render_value(field.load_value(review_request))

        self.assertHTMLEqual(
            result,
            """
            <div class="rb-c-review-request-field-tabular rb-c-commit-list">
             <table class="rb-c-review-request-field-tabular__data">
              <thead>
               <tr>
                <th class="rb-c-commit-list__column-summary">Summary</th>
                <th class="rb-c-commit-list__column-id">ID</th>
                <th class="rb-c-commit-list__column-author">Author</th>
               </tr>
              </thead>
              <tbody>
               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <div class="rb-c-commit-list__message-summary">Commit
                  message 1</div>
                </td>
                <td class="rb-c-commit-list__id" title="r1">r1</td>
                <td class="rb-c-commit-list__author">Example Author</td>
               </tr>

               <tr class="rb-c-commit-list__commit">
                <td class="rb-c-commit-list__message">
                 <details>
                  <summary class="rb-c-commit-list__message-summary">Commit
                   message 2</summary>
                  <div class="rb-c-commit-list__message-body">Longer
                   message</div>
                 </details>
                </td>
                <td class="rb-c-commit-list__id" title="r2">r2</td>
                <td class="rb-c-commit-list__author">Doc Dwarf</td>
               </tr>
              </tbody>
             </table>
            </div>
            """)

    def test_render_change_entry_html(self):
        """Testing CommitListField.render_change_entry_html"""
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        author_name = review_request.submitter.get_full_name()

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=author_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=author_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=author_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name=author_name)

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                </tr>
               </tbody>
              </table>
             </div>
            """)

    def test_render_change_entry_html_expand(self):
        """Testing CommitListField.render_change_entry_html with a multi-line
        commit message
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        author_name = review_request.submitter.get_full_name()

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1\n\n'
                                              'A long message.\n',
                               author_name=author_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=author_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=author_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2\n\n'
                                              'So very long of a message.\n',
                               author_name=author_name)

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <details>
                   <summary class="rb-c-commit-list__message-summary">Commit
                    message 1</summary>
                   <div class="rb-c-commit-list__message-body">A long
                    message.</div>
                  </details>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New commit
                   message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <details>
                   <summary class="rb-c-commit-list__message-summary">New
                    commit message 2</summary>
                   <div class="rb-c-commit-list__message-body">So very long
                    of a message.</div>
                  </details>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                </tr>
               </tbody>
              </table>
             </div>
            </div>
            """)

    def test_render_change_entry_html_expand_with_author(self):
        """Testing CommitListField.render_change_entry_html with an author that
        differs from the review request submitter and a multi-line commit
        message
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()
        self.assertEqual(submitter_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1\n\n'
                                              'A long message.\n',
                               author_name='Example Author')
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=submitter_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=submitter_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2\n\n'
                                              'So very long of a message.\n',
                               author_name=submitter_name)

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                 <th class="rb-c-commit-list__column-author">Author</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <details>
                   <summary class="rb-c-commit-list__message-summary">Commit
                    message 1</summary>
                   <div class="rb-c-commit-list__message-body">A long
                    message.</div>
                  </details>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Example Author</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New commit
                   message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <details>
                   <summary class="rb-c-commit-list__message-summary">New
                    commit message 2</summary>
                   <div class="rb-c-commit-list__message-body">So very long
                    of a message.</div>
                  </details>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>
               </tbody>
              </table>
             </div>
            </div>
            """)

    def test_render_change_entry_html_with_author_old(self):
        """Testing CommitListField.render_change_entry_html with an author that
        differs from the review request submitter in the old commits
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()
        self.assertEqual(submitter_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name='Example Author')
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=submitter_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=submitter_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name=submitter_name)

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                 <th class="rb-c-commit-list__column-author">Author</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Example Author</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>
               </tbody>
              </table>
             </div>
            </div>
            """)

    def test_render_change_entry_html_with_author_new(self):
        """Testing CommitListField.render_change_entry_html with an author that
        differs from the review request submitter in the new commits
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()
        self.assertEqual(submitter_name, 'Doc Dwarf')

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=submitter_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=submitter_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=submitter_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name='Example Author')

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                 <th class="rb-c-commit-list__column-author">Author</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-removed">
                 <td class="rb-c-commit-list__op"
                     aria-label="Removed commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">Commit
                   message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                 <td class="rb-c-commit-list__author">Doc Dwarf</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                 <td class="rb-c-commit-list__author">Example Author</td>
                </tr>
               </tbody>
              </table>
             </div>
            </div>
            """)

    def test_render_change_entry_html_first_diffset(self):
        """Testing CommitListfield.render_change_entry_html with a change that
        adds the first diffset
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        author_name = review_request.submitter.get_full_name()
        self.assertEqual(author_name, 'Doc Dwarf')

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=author_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name=author_name)

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)
        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertHTMLEqual(
            result,
            """
            <div class="commit-list-container">
             <div class="rb-c-review-request-field-tabular rb-c-commit-list">
              <table class="rb-c-review-request-field-tabular__data">
               <thead>
                <tr>
                 <th class="rb-c-commit-list__column-op"></th>
                 <th class="rb-c-commit-list__column-summary">Summary</th>
                 <th class="rb-c-commit-list__column-id">ID</th>
                </tr>
               </thead>
               <tbody>
                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 1</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r1">r1</td>
                </tr>

                <tr class="rb-c-commit-list__commit -is-added">
                 <td class="rb-c-commit-list__op"
                     aria-label="Added commit"></td>
                 <td class="rb-c-commit-list__message">
                  <div class="rb-c-commit-list__message-summary">New
                   commit message 2</div>
                 </td>
                 <td class="rb-c-commit-list__id" title="r2">r2</td>
                </tr>
               </tbody>
              </table>
             </div>
            </div>
            """)

    def test_serialize_change_entry(self):
        """Testing CommitListField.serialize_change_entry"""
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        diffset = self.create_diffset(review_request)

        submitter_name = review_request.submitter.get_full_name()

        self.create_diffcommit(diffset=diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='Commit message 1',
                               author_name=submitter_name)
        self.create_diffcommit(diffset=diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='Commit message 2',
                               author_name=submitter_name)

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=submitter_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name='Example Author')

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()
        field = self.make_field(review_request)

        self.assertEqual(
            {
                'old': [
                    {
                        'author': submitter_name,
                        'summary': 'Commit message 1',
                    },
                    {
                        'author': submitter_name,
                        'summary': 'Commit message 2',
                    },
                ],
                'new': [
                    {
                        'author': submitter_name,
                        'summary': 'New commit message 1',
                    },
                    {
                        'author': 'Example Author',
                        'summary': 'New commit message 2',
                    },
                ],
            },
            field.serialize_change_entry(changedesc))

    def serialize_change_entry_first_diffset(self):
        """Testing CommitListField.serialize_change_entry with a change that
        adds the first diffset
        """
        target = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository,
                                                    target_people=[target],
                                                    public=True,
                                                    create_with_history=True)
        submitter_name = review_request.submitter.get_full_name()

        draft_diffset = self.create_diffset(review_request, draft=True)
        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r1',
                               parent_id='r0',
                               commit_message='New commit message 1',
                               author_name=submitter_name)

        self.create_diffcommit(diffset=draft_diffset,
                               commit_id='r2',
                               parent_id='r1',
                               commit_message='New commit message 2',
                               author_name='Example Author')

        draft_diffset.finalize_commit_series(
            cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()
        field = self.make_field(review_request)

        self.assertEqual(
            {
                'old': None,
                'new': [
                    {
                        'author': submitter_name,
                        'summary': 'New commit message 1',
                    },
                    {
                        'author': 'Example Author',
                        'summary': 'New commit message 2',
                    },
                ],
            },
            field.serialize_change_entry(changedesc))


class FileAttachmentsFieldTests(FieldsTestCase):
    """Unit tests for FileAttachmentsField."""

    field_cls = FileAttachmentsField
    fixtures = ['test_users']

    def test_render_change_entry_html(self):
        """Testing FileAttachmentsField.render_change_entry_html"""
        target = User.objects.get(username='doc')
        review_request = self.create_review_request(public=True,
                                                    create_with_history=True,
                                                    target_people=[target])
        attachment1 = self.create_file_attachment(
            review_request,
            caption='Attachment 1',
            orig_filename='file1.png')

        attachment2 = self.create_file_attachment(
            review_request,
            draft=True,
            draft_caption='Attachment 2',
            orig_filename='file2.png')
        attachment3 = self.create_file_attachment(
            review_request,
            draft=True,
            draft_caption='Attachment 3',
            orig_filename='file3.png')

        draft = review_request.get_draft()
        draft.inactive_file_attachments.add(attachment1)
        draft.file_attachments.remove(attachment1)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)

        # 3 queries:
        #
        # 1. Fetch active file attachments
        # 2. Fetch inactive file attachments
        # 3. Fetch the review request draft
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_inactive_file_attachments',
                },
                'where': Q(inactive_review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
        ]

        # Check the added file attachments. Only file attachments 2 and 3
        # should be present.
        with assert_queries(queries):
            result = field.render_change_entry_html(
                changedesc.fields_changed[field.field_id]['added'])

        self.assertIsInstance(result, SafeText)

        self.assertNotIn('"id": %s,' % attachment1.pk, result)
        self.assertIn('"id": %s,' % attachment2.pk, result)
        self.assertIn('"id": %s,' % attachment3.pk, result)

        # Check the removed file attachments. Only file attachment 1
        # should be present.
        with self.assertNumQueries(0):
            result = field.render_change_entry_html(
                changedesc.fields_changed[field.field_id]['removed'])

        self.assertIsInstance(result, SafeText)

        self.assertIn('"id": %s,' % attachment1.pk, result)
        self.assertNotIn('"id": %s,' % attachment2.pk, result)
        self.assertNotIn('"id": %s,' % attachment3.pk, result)

    def test_get_change_entry_sections_html(self):
        """Testing FileAttachmentsField.get_change_entry_sections_html"""
        target = User.objects.get(username='doc')
        review_request = self.create_review_request(public=True,
                                                    create_with_history=True,
                                                    target_people=[target])
        attachment1 = self.create_file_attachment(
            review_request,
            caption='Attachment 1',
            orig_filename='file1.png')

        self.create_file_attachment(
            review_request,
            draft=True,
            draft_caption='Attachment 2',
            orig_filename='file2.png')
        self.create_file_attachment(
            review_request,
            draft=True,
            draft_caption='Attachment 3',
            orig_filename='file3.png')

        draft = review_request.get_draft()
        draft.inactive_file_attachments.add(attachment1)
        draft.file_attachments.remove(attachment1)

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)

        # 3 queries:
        #
        # 1. Fetch active file attachments
        # 2. Fetch inactive file attachments
        # 3. Fetch the review request draft
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_inactive_file_attachments',
                },
                'where': Q(inactive_review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
        ]

        # Check that the queries are only run once and that the cached data
        # is used subsequently when fetching the state of all file attachments
        # in the changedesc.
        with assert_queries(queries):
            field.get_change_entry_sections_html(
                changedesc.fields_changed[field.field_id])


class BugsFieldTests(FieldsTestCase):
    """Unit tests for BugsField.

    Version Added:
        8.0
    """

    field_cls = BugsField
    fixtures = ['test_users']

    def test_render_change_entry_html_with_added_and_removed(self) -> None:
        """Testing BugsField.render_change_entry_html with bugs added and
        removed
        """
        review_request = self.create_review_request(
            public=True,
            bugs_closed='123,456',
            target_people=[User.objects.get(username='doc')])

        draft = ReviewRequestDraft.create(review_request)
        draft.bugs_closed = '456,789'
        draft.save()

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)

        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertIsInstance(result, SafeText)
        self.assertIn('123', result)
        self.assertIn('789', result)

        # Make sure bugs are rendered as plain text, not as Python list
        # representations (e.g., "['123']").
        self.assertNotIn('[', result)
        self.assertNotIn(']', result)

    def test_render_change_entry_html_with_only_added(self) -> None:
        """Testing BugsField.render_change_entry_html with only bugs added"""
        review_request = self.create_review_request(
            public=True,
            bugs_closed='123',
            target_people=[User.objects.get(username='doc')])

        draft = ReviewRequestDraft.create(review_request)
        draft.bugs_closed = '123,456'
        draft.save()

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)

        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertIsInstance(result, SafeText)
        self.assertIn('456', result)
        self.assertNotIn('[', result)
        self.assertNotIn(']', result)

    def test_render_change_entry_html_with_only_removed(self) -> None:
        """Testing BugsField.render_change_entry_html with only bugs
        removed
        """
        review_request = self.create_review_request(
            public=True,
            bugs_closed='123,456',
            target_people=[User.objects.get(username='doc')])

        draft = ReviewRequestDraft.create(review_request)
        draft.bugs_closed = '123'
        draft.save()

        review_request.publish(user=review_request.submitter)
        changedesc = review_request.changedescs.latest()

        field = self.make_field(review_request)

        result = field.render_change_entry_html(
            changedesc.fields_changed[field.field_id])

        self.assertIsInstance(result, SafeText)
        self.assertIn('456', result)
        self.assertNotIn('[', result)
        self.assertNotIn(']', result)



"""Tests for reviewboard.reviews.signal_handlers.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Q, Value

from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.diffviewer.models import DiffSet, DiffSetHistory, FileDiff
from reviewboard.diffviewer.models.diffcommit import DiffCommit
from reviewboard.reviews.models import (Comment,
                                        FileAttachmentComment,
                                        Review,
                                        ReviewRequest,
                                        ReviewRequestDraft,
                                        Screenshot,
                                        StatusUpdate)
from reviewboard.reviews.signal_handlers import (
    _on_review_request_deleted,
    _on_review_request_draft_deleted)
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class OnReviewRequestDraftDeletedTests(kgb.SpyAgency, TestCase):
    """Tests _on_review_request_draft_deleted() signal handler.

    Version Added:
        6.0
    """

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()
        self.review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        self.draft = self.create_review_request_draft(self.review_request)

    def test_with_file_attachments(self) -> None:
        """Testing _on_review_request_draft_deleted deletes new and new
        revision draft file attachments
        """
        self.spy_on(_on_review_request_draft_deleted)

        review_request = self.review_request
        draft = self.draft
        changedesc = draft.changedesc

        assert draft is not None
        assert changedesc is not None

        published_file_attachment = self.create_file_attachment(
            review_request,
            caption='Published')
        new_file_attachment = self.create_file_attachment(
            review_request,
            caption='New',
            draft=True)
        new_revision_file_attachment = self.create_file_attachment(
            review_request,
            attachment_history=published_file_attachment.attachment_history,
            attachment_revision=(
                published_file_attachment.attachment_revision + 1),
            caption='New Revision',
            draft=True)

        # Clear the review request's caches so we'll re-fetch the draft.
        review_request.clear_local_caches()

        queries: ExpectedQueries = [
            {
                'model': ReviewRequestDraft.target_groups.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.target_people.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.depends_on.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(drafts__id=1),
            },
            {
                'model': FileAttachmentHistory,
                'where': Q(id=2),
            },
            {
                'model': FileAttachmentHistory,
                'where': Q(id=1),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=1),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(file_attachment__in=[new_revision_file_attachment]),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(diff_against_file_attachment__in=[
                    new_revision_file_attachment,
                ]),
            },
            {
                'model': ReviewRequest.file_attachments.through,
                'where': Q(fileattachment__in=[new_revision_file_attachment]),
            },
            {
                'model': ReviewRequest.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_revision_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(fileattachment__in=[new_revision_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_revision_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[new_file_attachment.pk]),
            },
            {
                'model': FileAttachment,
                'type': 'DELETE',
                'where': Q(id__in=[new_revision_file_attachment.pk]),
            },
            {
                'model': FileAttachmentHistory,
                'type': 'UPDATE',
                'where': Q(pk=published_file_attachment.pk),
            },
            {
                'limit': 1,
                'model': FileAttachmentHistory,
                'values_select': ('latest_revision',),
                'where': Q(pk=published_file_attachment.pk),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(file_attachment__in=[new_file_attachment]),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(diff_against_file_attachment__in=[
                    new_file_attachment,
                ]),
            },
            {
                'model': ReviewRequest.file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequest.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[published_file_attachment.pk]),
            },
            {
                'model': FileAttachment,
                'type': 'DELETE',
                'where': Q(id__in=[new_file_attachment.pk]),
            },
            {
                'model': FileAttachmentHistory,
                'type': 'UPDATE',
                'where': Q(pk=new_file_attachment.pk),
            },
            {
                'limit': 1,
                'model': FileAttachmentHistory,
                'values_select': ('latest_revision',),
                'where': Q(pk=new_file_attachment.pk),
            },
            {
                'annotations': {
                    'a': Value(1),
                },
                'join_types': {
                    'reviews_reviewrequest_changedescs': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_changedescs',
                },
                'where': Q(changedescs__id=changedesc.pk),
            },
            {
                'model': ReviewRequest.changedescs.through,
                'where': Q(changedescription__in=[changedesc]),
            },
            {
                'model': StatusUpdate,
                'where': Q(change_description__in=[changedesc]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(changedesc__in=[changedesc]),
            },
            {
                'model': ChangeDescription,
                'type': 'DELETE',
                'where': Q(id__in=[changedesc.pk]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[
                    new_file_attachment.pk,
                    published_file_attachment.pk,
                ]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'DELETE',
                'where': Q(id__in=[draft.pk]),
            },
        ]

        with assert_queries(queries):
            self.draft.delete()

        all_attachments = list(FileAttachment.objects.all())

        self.assertSpyCalled(_on_review_request_draft_deleted)
        self.assertNotIn(new_file_attachment, all_attachments)
        self.assertNotIn(new_revision_file_attachment, all_attachments)
        self.assertIn(published_file_attachment, all_attachments)

    def test_with_one_file_attachment(self) -> None:
        """Testing _on_review_request_draft_deleted deletes a new file
        attachment
        """
        self.spy_on(_on_review_request_draft_deleted)

        review_request = self.review_request
        draft = self.draft
        changedesc = draft.changedesc

        assert draft is not None
        assert changedesc is not None

        published_file_attachment = self.create_file_attachment(
            review_request,
            caption='Published')
        new_file_attachment = self.create_file_attachment(
            review_request,
            draft=True,
            caption='New')

        # Clear the review request's caches so we'll re-fetch the draft.
        review_request.clear_local_caches()

        queries: ExpectedQueries = [
            {
                'model': ReviewRequestDraft.target_groups.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.target_people.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.depends_on.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(drafts__id=1),
            },
            {
                'model': FileAttachmentHistory,
                'where': Q(id=2),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=1),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(file_attachment__in=[new_file_attachment]),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(diff_against_file_attachment__in=[
                    new_file_attachment,
                ]),
            },
            {
                'model': ReviewRequest.file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequest.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[new_file_attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[published_file_attachment.pk]),
            },
            {
                'model': FileAttachment,
                'type': 'DELETE',
                'where': Q(id__in=[new_file_attachment.pk]),
            },
            {
                'model': FileAttachmentHistory,
                'type': 'UPDATE',
                'where': Q(pk=new_file_attachment.pk),
            },
            {
                'limit': 1,
                'model': FileAttachmentHistory,
                'values_select': ('latest_revision',),
                'where': Q(pk=new_file_attachment.pk),
            },
            {
                'annotations': {
                    'a': Value(1),
                },
                'join_types': {
                    'reviews_reviewrequest_changedescs': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_changedescs',
                },
                'where': Q(changedescs__id=changedesc.pk),
            },
            {
                'model': ReviewRequest.changedescs.through,
                'where': Q(changedescription__in=[changedesc]),
            },
            {
                'model': StatusUpdate,
                'where': Q(change_description__in=[changedesc]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(changedesc__in=[changedesc]),
            },
            {
                'model': ChangeDescription,
                'type': 'DELETE',
                'where': Q(id__in=[changedesc.pk]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[published_file_attachment.pk]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'DELETE',
                'where': Q(id__in=[draft.pk]),
            },
        ]

        with assert_queries(queries, with_tracebacks=True):
            draft.delete()

        all_attachments = list(FileAttachment.objects.all())

        self.assertSpyCalled(_on_review_request_draft_deleted)
        self.assertNotIn(new_file_attachment, all_attachments)
        self.assertIn(published_file_attachment, all_attachments)

    def test_with_no_file_attachments(self) -> None:
        """Testing _on_review_request_draft_deleted when there's no
        draft file attachments on the review request
        """
        self.spy_on(_on_review_request_draft_deleted)

        # 13 queries:
        #
        #   1-7. Fetch review request draft info and relations
        #  8-12. Delete the change description
        #    13. Delete the review request draft
        with self.assertNumQueries(13):
            self.draft.delete()

        self.assertSpyCalled(_on_review_request_draft_deleted)

    def test_with_change_description(self) -> None:
        """Testing _on_review_request_draft_deleted with a change description
        """
        self.spy_on(_on_review_request_draft_deleted)
        changedesc = ChangeDescription.objects.create(
            user=User.objects.get(username='doc'),
            public=False,
            text='x')

        self.draft.changedesc = changedesc
        self.draft.save()

        # 13 queries:
        #
        #  1-7. Fetch review request draft info and relations
        # 8-12. Delete the change description
        #   13. Delete the review request draft
        with self.assertNumQueries(13):
            self.draft.delete()

        self.assertNotIn(changedesc, ChangeDescription.objects.all())

    def test_with_diff_and_diff_file_attachment(self) -> None:
        """Testing _on_review_request_draft_deleted with a diff and diff file
        attachment
        """
        self.spy_on(_on_review_request_draft_deleted)

        review_request = self.review_request
        draft = self.draft
        changedesc = draft.changedesc

        assert draft is not None
        assert changedesc is not None

        diffset = self.create_diffset(
            repository=review_request.repository)
        draft.diffset = diffset
        draft.save(update_fields=('diffset',))

        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        # Clear the review request's caches so we'll re-fetch the draft.
        review_request.clear_local_caches()

        queries: ExpectedQueries = [
            {
                'model': ReviewRequestDraft.target_groups.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.target_people.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_screenshots.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'model': ReviewRequestDraft.depends_on.through,
                'where': Q(reviewrequestdraft__in=[draft]),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(drafts__id=1),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=1),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'annotations': {
                    'a': Value(1),
                },
                'join_types': {
                    'reviews_reviewrequest_changedescs': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_changedescs',
                },
                'where': Q(changedescs__id=changedesc.pk),
            },
            {
                'model': ReviewRequest.changedescs.through,
                'where': Q(changedescription__in=[changedesc]),
            },
            {
                'model': StatusUpdate,
                'where': Q(change_description__in=[changedesc]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(changedesc__in=[changedesc]),
            },
            {
                'model': ChangeDescription,
                'type': 'DELETE',
                'where': Q(id__in=[changedesc.pk]),
            },
            {
                'model': DiffCommit,
                'where': Q(diffset__in=[diffset]),
            },
            {
                'model': FileDiff,
                'where': Q(diffset__in=[diffset]),
            },
            {
                'model': FileAttachment,
                'where': Q(added_in_filediff__in=[filediff]),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(file_attachment__in=[attachment]),
            },
            {
                'model': FileAttachmentComment,
                'where': Q(diff_against_file_attachment__in=[attachment]),
            },
            {
                'model': ReviewRequest.file_attachments.through,
                'where': Q(fileattachment__in=[attachment]),
            },
            {
                'model': ReviewRequest.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[attachment]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'where': Q(fileattachment__in=[attachment]),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': Q(fileattachment__in=[attachment]),
            },
            {
                'model': Comment,
                'where': Q(filediff__in=[filediff]),
            },
            {
                'model': Comment,
                'where': Q(interfilediff__in=[filediff]),
            },
            {
                'model': Review,
                'where': Q(reviewed_diffset__in=[diffset]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(diffset__in=[diffset]),
            },
            {
                'model': FileDiff,
                'type': 'DELETE',
                'where': Q(id__in=[filediff.pk]),
            },
            {
                'model': ReviewRequest.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[attachment.pk]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[attachment.pk]),
            },
            {
                'model': DiffSet,
                'type': 'DELETE',
                'where': Q(id__in=[diffset.pk]),
            },
            {
                'model': FileAttachment,
                'type': 'DELETE',
                'where': Q(id__in=[attachment.pk]),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[attachment.pk]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'DELETE',
                'where': Q(id__in=[draft.pk]),
            },
        ]

        with assert_queries(queries):
            draft.delete()

        self.assertNotIn(attachment, FileAttachment.objects.all())
        self.assertNotIn(filediff, FileDiff.objects.all())
        self.assertNotIn(diffset, DiffSet.objects.all())

    def test_handler_does_not_delete_published_data(self) -> None:
        """Testing that _on_review_request_draft_deleted running after a draft
        publish does not delete data that is now part of the review request
        """
        self.spy_on(_on_review_request_draft_deleted)

        user = User.objects.get(username='doc')

        diffset = self.create_diffset(
            repository=self.review_request.repository)
        self.draft.diffset = diffset

        changedesc = ChangeDescription.objects.create(
            user=user,
            public=False,
            text='x')
        self.draft.changedesc = changedesc

        attachment = self.create_file_attachment(self.review_request,
                                                 draft=True)

        self.draft.save()

        self.draft.target_people.add(user)
        self.review_request.publish(user)

        self.assertIn(diffset, DiffSet.objects.all())
        self.assertIn(changedesc, ChangeDescription.objects.all())
        self.assertTrue(attachment, FileAttachment.objects.all())


class OnReviewRequestDeletedTests(kgb.SpyAgency, TestCase):
    """Tests for the _on_review_request_deleted signal handler.

    Version Added:
        7.0
    """

    fixtures = ['test_scmtools', 'test_users']

    def test_cleans_up_related_changedescs(self) -> None:
        """Testing _on_review_request_deleted deletes related ChangeDescription
        objects
        """
        self.spy_on(_on_review_request_deleted)

        review_request = self.create_review_request(
            publish=True)

        c = ChangeDescription.objects.create(
            user=User.objects.get(username='doc'),
            public=True,
            text='x')
        review_request.changedescs.add(c)

        self.assertEqual(ChangeDescription.objects.count(), 1)

        # 36 queries:
        #
        #  1-11. Update profiles and counts.
        # 12-29. Fetch review request and related object data.
        # 30-31. Delete change description and relation.
        # 32-35. Delete other related data.
        #    36. Delete review request.
        with self.assertNumQueries(36):
            review_request.delete()

        self.assertSpyCalled(_on_review_request_deleted)
        self.assertEqual(ChangeDescription.objects.count(), 0)

    def test_cleans_up_related_diffset_history(self) -> None:
        """Testing _on_review_request_deleted deletes related DiffSetHistory
        object
        """
        self.spy_on(_on_review_request_deleted)

        review_request = self.create_review_request(
            create_repository=True,
            publish=True)

        self.create_diffset(review_request)
        self.create_diffset(review_request, revision=2)

        self.assertEqual(DiffSetHistory.objects.count(), 1)

        # 35 queries:
        #
        #  1-11. Update profiles and counts.
        # 12-31. Fetch review request and related object data.
        #    32. Set diffset_history relation to NULL.
        #    33. Delete diffset history.
        #    34. Delete diffsets.
        #    35. Delete review request.
        with self.assertNumQueries(35):
            review_request.delete()

        self.assertSpyCalled(_on_review_request_deleted)
        self.assertEqual(DiffSetHistory.objects.count(), 0)

    def test_cleans_up_related_file_attachments(self) -> None:
        """Testing _on_review_request_deleted deletes related file attachments
        """
        self.spy_on(_on_review_request_deleted)

        review_request = self.create_review_request(
            create_repository=True,
            publish=True)

        self.create_file_attachment(review_request, active=True)
        self.create_file_attachment(review_request, active=False)

        self.assertEqual(FileAttachment.objects.count(), 2)
        self.assertEqual(FileAttachmentHistory.objects.count(), 2)
        self.assertEqual(review_request.file_attachments_count, 1)
        self.assertEqual(review_request.inactive_file_attachments_count, 1)

        # 51 queries:
        #
        #  1-11. Update profiles and counts.
        # 12-35. Fetch review request and related object data.
        # 36-39. Remove file attachment and file attachment history relations.
        #    40. Delete file attachments.
        # 41-44. Perform file attachment relation bookkeeping.
        # 45-47. Delete diffset history.
        # 48-50. Clean up file attachment relations.
        #    51. Delete review request.
        with self.assertNumQueries(51):
            review_request.delete()

        self.assertSpyCalled(_on_review_request_deleted)
        self.assertEqual(FileAttachment.objects.count(), 0)
        self.assertEqual(FileAttachmentHistory.objects.count(), 0)

    def test_cleans_up_related_screenshots(self) -> None:
        """Testing _on_review_request_deleted deletes related screenshots"""
        self.spy_on(_on_review_request_deleted)

        review_request = self.create_review_request(
            create_repository=True,
            publish=True)

        self.create_screenshot(review_request, active=True)
        self.create_screenshot(review_request, active=False)

        self.assertEqual(Screenshot.objects.count(), 2)
        self.assertEqual(review_request.screenshots_count, 1)
        self.assertEqual(review_request.inactive_screenshots_count, 1)

        # 51 queries:
        #
        #  1-11. Update profiles and counts.
        # 12-35. Fetch review request and related object data.
        # 36-41. Remove screenshot relations.
        #    42. Delete screenshots.
        # 43-45. Delete diffset history.
        # 46-47. Clean up screenshot relations..
        #    48. Delete review request.
        with self.assertNumQueries(48):
            review_request.delete()

        self.assertSpyCalled(_on_review_request_deleted)
        self.assertEqual(Screenshot.objects.count(), 0)



"""E-mail notification callbacks."""

from typing import Union

from django.contrib.auth.models import User
from django.http import HttpRequest
from django.utils import timezone
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.changedescs.models import ChangeDescription
from reviewboard.notifications.email.message import (
    prepare_password_changed_mail,
    prepare_reply_published_mail,
    prepare_review_published_mail,
    prepare_review_request_mail,
    prepare_user_registered_mail,
    prepare_webapi_token_mail)
from reviewboard.notifications.email.utils import send_email
from reviewboard.reviews.models import Review, ReviewRequest
from reviewboard.webapi.models import WebAPIToken


def _update_email_info(
    obj: Union[Review, ReviewRequest],
    message_id: str,
) -> None:
    """Update the e-mail message information on the object.

    The ``email_message_id`` and ``time_emailed`` fields of the model will be
    updated.

    Args:
        obj (reviewboard.reviews.models.review.Review or
             reviewboard.reviews.models.review_request.ReviewRequest):
            The object for which e-mail information will be updated.

        message_id (str):
            The new e-mail message ID.
    """
    obj.email_message_id = message_id
    obj.time_emailed = timezone.now()
    obj.save(update_fields=('email_message_id', 'time_emailed'))


def send_password_changed_mail(user: User) -> None:
    """Send an e-mail when a user's password changes.

    The e-mail will only be sent when the ``mail_send_password_changed_mail``
    :py:class:`~djblets.siteconfig.models.SiteConfiguration` setting is
    enabled.

    Args:
        user (django.contrib.auth.model.User):
            The user whose password changed.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if siteconfig.get('mail_send_password_changed_mail'):
        send_email(prepare_password_changed_mail, user=user)


def send_reply_published_mail(
    user: User,
    reply: Review,
    trivial: bool,
    **kwargs,
) -> None:
    """Send e-mail when a review reply is published.

    Listens to the :py:data:`~reviewboard.reviews.signals.reply_published`
    signal and sends an e-mail if this type of notification is enabled (through
    ``mail_send_review_mail`` site configuration).

    Args:
        user (django.contrib.auth.models.User):
            The user who published the reply.

        reply (reviewboard.reviews.models.Review):
            The reply that was published.

        trivial (bool):
            Whether or not the reply was marked as trivial when it was
            published. If ``True``, the e-mail will not be sent.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if trivial or not siteconfig.get('mail_send_review_mail'):
        return

    review_request = reply.review_request

    if not review_request.public:
        return

    review = reply.base_reply_to

    message, sent = send_email(prepare_reply_published_mail,
                               user=user,
                               reply=reply,
                               review=review,
                               review_request=review_request)

    if sent:
        assert message is not None
        _update_email_info(reply, message.message_id)


def send_review_published_mail(
    user: User,
    review: Review,
    to_owner_only: bool,
    trivial: bool,
    request: HttpRequest,
    **kwargs,
) -> None:
    """Send e-mail when a review is published.

    Listens to the :py:data:`~reviewboard.reviews.signals.review_published`
    signal and sends e-mail if this type of notification is enabled (through
    the ``mail_send_review_mail`` site configuration setting).

    Args:
        user (django.contrib.auth.models.User):
            The user that published the review.

        review (reviewboard.reviews.models.review.Review):
            The review that was published.

        to_owner_only (bool):
            Whether or not the mail should only be sent to the review request
            submitter.

        trivial (bool):
            Whether to skip e-mail notifications.

        request (django.http.HttpRequest):
            The HTTP request that triggered this e-mail.

        **kwargs (dict):
            Additional keyword arguments.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if trivial or not siteconfig.get('mail_send_review_mail'):
        return

    review_request = review.review_request

    if not review_request.public:
        return

    message, sent = send_email(prepare_review_published_mail,
                               user=user,
                               review=review,
                               review_request=review_request,
                               request=request,
                               to_owner_only=to_owner_only)

    if sent:
        assert message is not None
        _update_email_info(review, message.message_id)


def send_review_request_closed_mail(
    user: User,
    review_request: ReviewRequest,
    close_type: str,
    **kwargs,
) -> None:
    """Send e-mail when a review request is closed.

    Listens to the
    :py:data:`~reviewboard.reviews.signals.review_request_closed` signal and
    sends an e-mail if this type of notification is enabled (through the
    ``mail_send_review_close_mail`` site configuration setting).

    Args:
        user (django.contrib.auth.models.User):
            The user who closed the review request.

        review_request (reviewboard.reviews.models.review_request.
                        ReviewRequest):
            The review request that was closed.

        close_type (str):
            How the review request was closed.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not (siteconfig.get('mail_send_review_close_mail') and
            review_request.public):
        return

    message, sent = send_email(prepare_review_request_mail,
                               user=user,
                               review_request=review_request,
                               close_type=close_type)

    if sent:
        assert message is not None
        _update_email_info(review_request, message.message_id)


def send_review_request_published_mail(
    user: User,
    review_request: ReviewRequest,
    trivial: bool,
    changedesc: ChangeDescription,
    **kwargs,
) -> None:
    """Send e-mail when a review request is published.

    Listens to the
    :py:data:`~reviewboard.reviews.signals.review_request_published` signal and
    sends an e-mail if this type of notification is enabled (through the
    ``mail_send_review_mail`` site configuration setting).

    If the review request publish was marked as trivial, the e-mail will not be
    sent.

    Args:
        user (django.contrib.auth.models.User):
            The user who published the review request.

        review_request (reviewboard.reviews.models.review_request.
                        ReviewRequest):
            The review request that was published.

        trivial (bool):
            Whether or not the publish was marked as trivial.

        changedesc (reviewboard.changedescs.models.ChangeDescription,
                    optional):
            The change description associated with the publish, if any.

        **kwargs (dict):
            Ignored keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if trivial or not siteconfig.get('mail_send_review_mail'):
        return

    if (not review_request.public or
        review_request.status == ReviewRequest.DISCARDED):
        return

    message, sent = send_email(prepare_review_request_mail,
                               user=user,
                               review_request=review_request,
                               changedesc=changedesc)

    if sent:
        assert message is not None
        _update_email_info(review_request, message.message_id)


def send_user_registered_mail(
    user: User,
    **kwargs,
) -> None:
    """Send e-mail when a user is registered.

    Listens for new user registrations and sends a new user registration
    e-mail to administrators, if this type of notification is enabled (through
    ``mail_send_new_user_mail`` site configuration).

    Args:
        user (django.contrib.auth.models.User):
            The user that registered.

        **kwargs (dict):
            Unused keyword arguments from the signal.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not siteconfig.get('mail_send_new_user_mail'):
        return

    send_email(prepare_user_registered_mail,
               user=user)


def send_webapi_token_created_mail(
    instance: WebAPIToken,
    auto_generated: bool = False,
    **kwargs,
) -> None:
    """Send e-mail when an API token is created.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that has been created.

        auto_generated (bool, optional):
            Whether or not the token is being automatically generated. If it
            is automatically generated then no email will be sent.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    if not auto_generated:
        send_email(prepare_webapi_token_mail,
                   webapi_token=instance,
                   op='created')


def send_webapi_token_deleted_mail(
    instance: WebAPIToken,
    **kwargs,
) -> None:
    """Send e-mail when an API token is deleted.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that has been deleted.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    send_email(prepare_webapi_token_mail,
               webapi_token=instance,
               op='deleted')


def send_webapi_token_expired_mail(
    instance: WebAPIToken,
    **kwargs,
) -> None:
    """Send e-mail upon the first use of an API token that is expired.

    Version Added:
        5.0

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that is expired.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    if not instance.extra_data.get('expired_notification_sent'):
        sent = send_email(prepare_webapi_token_mail,
                          webapi_token=instance,
                          op='expired')[1]

        if sent:
            instance.extra_data['expired_notification_sent'] = True
            instance.save_base(update_fields=('extra_data',))


def send_webapi_token_updated_mail(
    instance: WebAPIToken,
    **kwargs,
) -> None:
    """Send e-mail when an API token is updated.

    Args:
        instance (reviewboard.webapi.models.WebAPIToken):
            The token that was updated.

        **kwargs (dict):
            Unused keyword arguments provided by the signal.
    """
    send_email(prepare_webapi_token_mail,
               webapi_token=instance,
               op='updated')



"""Testing utilities for building expected queries for review groups.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Dict, Set, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Q

from reviewboard.accounts.testing.queries import (
    get_user_local_site_profile_equeries,
    get_user_permissions_equeries,
)
from reviewboard.reviews.models import Group
from reviewboard.site.models import LocalSite
from reviewboard.site.testing.queries import \
    get_local_site_is_mutable_by_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries

    from reviewboard.site.models import AnyOrAllLocalSites
    from reviewboard.testing.queries.base import ExpectedQResult


def get_review_groups_accessible_q(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    visible_only: bool = True,
    has_view_invite_only_groups_perm: bool = False,
    needs_local_site_profile_query: bool = False,
    needs_user_permission_queries: bool = True,
    users_join_type: str = 'INNER JOIN',
) -> ExpectedQResult:
    """Return a Q expression for accessible review group queries.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        visible_only (bool, optional):
            Whether the query is limited to visible review groups.

        has_view_invite_only_groups_perm (bool, optional):
            Whether to expect the ``reviews.has_view_invite_only_groups_perm``
            permission to be set.

            Set to ``True`` if this is not cached at this point.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set to ``True`` if this is not cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

        users_join_type (str, optional):
            The join type expected for any users relations.

            This defaults to `INNER JOIN`, as that's more likely in cases where
            this is being embedded in another query, outside of the manager's
            own methods.

    Returns:
        dict:
        The expected Q results.
    """
    tables: Set[str] = {'reviews_group'}
    join_types: Dict[str, str] = {}

    if user.is_superuser:
        if visible_only:
            if local_site is LocalSite.ALL:
                q = Q(visible=True)
            else:
                q = (Q(visible=True) &
                     Q(local_site=local_site))
        else:
            if local_site is LocalSite.ALL:
                q = Q()
            else:
                q = Q(local_site=local_site)
    elif user.is_authenticated:
        if has_view_invite_only_groups_perm and visible_only:
            tables.add('reviews_group_users')
            join_types['reviews_group_users'] = users_join_type

            if local_site is LocalSite.ALL:
                q = (Q(visible=True) |
                     Q(users=user.pk))
            else:
                q = ((Q(visible=True) |
                      Q(users=user.pk)) &
                     Q(local_site=local_site))
        elif has_view_invite_only_groups_perm and not visible_only:
            if local_site is LocalSite.ALL:
                q = Q()
            else:
                q = Q(local_site=local_site)
        elif not has_view_invite_only_groups_perm and visible_only:
            tables.add('reviews_group_users')
            join_types['reviews_group_users'] = users_join_type

            if local_site is LocalSite.ALL:
                q = ((Q(invite_only=False) &
                      Q(visible=True)) |
                     Q(users=user.pk))
            else:
                q = (((Q(invite_only=False) &
                       Q(visible=True)) |
                      Q(users=user.pk)) &
                     Q(local_site=local_site))
        else:
            tables.add('reviews_group_users')
            join_types['reviews_group_users'] = users_join_type

            if local_site is LocalSite.ALL:
                q = (Q(invite_only=False) |
                     Q(users=user.pk))
            else:
                q = ((Q(invite_only=False) |
                      Q(users=user.pk)) &
                     Q(local_site=local_site))
    else:
        assert user.is_anonymous

        if visible_only:
            if local_site is LocalSite.ALL:
                q = (Q(invite_only=False) &
                     Q(visible=True))
            else:
                q = ((Q(invite_only=False) &
                      Q(visible=True)) &
                     Q(local_site=local_site))
        else:
            if local_site is LocalSite.ALL:
                q = Q(invite_only=False)
            else:
                q = (Q(invite_only=False) &
                     Q(local_site=local_site))

    return {
        'join_types': join_types,
        'prep_equeries': get_review_groups_accessible_prep_equeries(
            user=user,
            local_site=local_site,
            needs_local_site_profile_query=needs_local_site_profile_query,
            needs_user_permission_queries=needs_user_permission_queries),
        'q': q,
        'tables': tables,
    }


def get_review_groups_accessible_prep_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    needs_local_site_profile_query: bool = False,
    needs_user_permission_queries: bool = True,
) -> ExpectedQueries:
    """Return expected Group accessibility preparation queries.

    These are queries that must be performed before the main accessibility
    query is executed.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set to ``True`` if this is not cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

    Returns:
        list of dict:
        The list of expected queries.
    """
    equeries: ExpectedQueries = []

    if user.is_authenticated and not user.is_superuser:
        assert isinstance(user, User)

        # This starts out by checking for user permissions, both on the
        # User/Group and on the Local Site.
        if needs_user_permission_queries:
            equeries += get_user_permissions_equeries(user=user)

        if local_site is not None and local_site is not LocalSite.ALL:
            equeries += get_local_site_is_mutable_by_equeries(
                local_site=local_site,
                user=user)

            if needs_local_site_profile_query:
                equeries += get_user_local_site_profile_equeries(
                    user=user,
                    profile=user.get_profile(),
                    local_site=local_site)

    return equeries


def get_review_groups_accessible_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    visible_only: bool = True,
    has_view_invite_only_groups_perm: bool = False,
    needs_local_site_profile_query: bool = False,
    needs_user_permission_queries: bool = True,
) -> ExpectedQueries:
    """Return Group accessibility preparation queries.

    This corresponds to a call from
    :py:meth:`Group.objects.accessible()
    <reviewboard.reviews.managers.GroupManager.accessible>`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        visible_only (bool, optional):
            Whether the queries are for visible groups only.

        has_view_invite_only_groups_perm (bool, optional):
            Whether to expect the ``reviews.has_view_invite_only_groups_perm``
            permission to be set.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set to ``False`` if this should be cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_groups_accessible_q(
        user=user,
        local_site=local_site,
        visible_only=visible_only,
        has_view_invite_only_groups_perm=has_view_invite_only_groups_perm,
        users_join_type='LEFT OUTER JOIN',
        needs_local_site_profile_query=needs_local_site_profile_query,
        needs_user_permission_queries=needs_user_permission_queries)
    q_tables = q_result['tables']

    equeries: ExpectedQueries = q_result.get('prep_equeries', []) + [
        {
            '__note__': 'Fetch a list of accessible review groups',
            'distinct': True,
            'join_types': q_result.get('join_types', {}),
            'model': Group,
            'num_joins': len(q_tables) - 1,
            'tables': q_tables,
            'where': q_result['q'],
        },
    ]

    return equeries


def get_review_groups_accessible_ids_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    visible_only: bool = True,
    has_view_invite_only_groups_perm: bool = False,
    needs_local_site_profile_query: bool = False,
    needs_user_permission_queries: bool = True,
) -> ExpectedQueries:
    """Return expected queries for Group.objects.accessible_ids().

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        visible_only (bool, optional):
            Whether the queries are for visible groups only.

        has_view_invite_only_groups_perm (bool, optional):
            Whether to expect the ``reviews.has_view_invite_only_groups_perm``
            permission to be set.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set to ``False`` if this should be cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

    Returns:
        list of dict:
        The list of expected queries.
    """
    equeries = get_review_groups_accessible_equeries(
        user=user,
        local_site=local_site,
        visible_only=visible_only,
        has_view_invite_only_groups_perm=has_view_invite_only_groups_perm,
        needs_local_site_profile_query=needs_local_site_profile_query,
        needs_user_permission_queries=needs_user_permission_queries)
    equeries[-1].update({
        '__note__': 'Fetch a list of accessible review group IDs',
        'distinct': False,
        'values_select': ('pk',),
    })

    return equeries
